<?php

namespace Drupal\server_beacon\Drush\Commands;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\server_beacon\ReporterInterface;
use Drush\Attributes\Argument;
use Drush\Attributes\Command;
use Drush\Attributes\Option;
use Drush\Attributes\Usage;
use Drush\Commands\DrushCommands;

/**
 * Drush commands for server beacon reports.
 */
class ServerReportCommands extends DrushCommands {

  /**
   * Creates a new instance of the ServiceReportCommands Drush commands object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\server_beacon\ReporterInterface $reporter
   *   The beacon reporter service for building and transmitting reports.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected ReporterInterface $reporter,
  ) {}

  /**
   * Retrieve a list of available server beacon reports.
   */
  #[Command(name: 'server-beacon:list', aliases: ['beaconls'])]
  #[Usage(
    name: 'drush server-beacon:list ',
    description: 'Get a list of configured available server reports.',
  )]
  public function list(): void {
    $reportStorage = $this->entityTypeManager->getStorage('server_beacon_report');

    /** @var \Drupal\server_beacon\Entity\ServerBeaconReportInterface[] $reports */
    $reports = $reportStorage->loadMultiple();
    if ($reports) {
      $headers = ['Name', 'Report to', 'Types'];
      $reportRows = [];
      foreach ($reports as $report) {
        $reportRows[] = [
          $report->id(),
          $report->reportHost(),
          implode(', ', $report->getBeaconTypes()),
        ];
      }

      $this->io()->table($headers, $reportRows);
    }
    else {
      $this->logger()->notice('No reports available.');
    }
  }

  /**
   * Display a generated report on the terminal.
   *
   * Helpful for locally viewing a local report result and ensuring that
   * report generation values appear as expected.
   *
   * @param string $report
   *   The service report identifier to send.
   */
  #[Command(name: 'server-beacon:print', aliases: ['beaconpr'])]
  #[Argument(name: 'report', description: 'The name of the report to display')]
  #[Usage(
    name: 'drush server-beacon:report <report>',
    description: 'Should the display the resulting report data in JSON format.',
  )]
  public function report(string $report): void {
    if ($data = $this->reporter->generateReport($report)) {
      $json = json_encode($data, JSON_PRETTY_PRINT);
      $this->output->write($json, TRUE);
    }
    else {
      $this->logger()->error('Failed to generate report data.');
    }
  }

  /**
   * Generates and transmits a server report to a server station.
   *
   * The server station is defined as the "reportTo" property of the
   * server_beacon_report config entity.
   *
   * @param string $report
   *   The service report identifier to send.
   * @param string[] $options
   *   Additional command options to apply when running this command.
   */
  #[Command(name: 'server-beacon:transmit', aliases: ['beacontx'])]
  #[Argument(name: 'report', description: 'The name of the report or project to transmit')]
  #[Option(
    name: 'environment',
    description: 'An environment identifier for the instance that is reporting',
    suggestedValues: ['qa', 'stage', 'prod'],
  )]
  #[Usage(
    name: 'drush server-beacon:transmit [--environment=<environment-id>] <report>',
    description: 'Transmit a report to the designated server station.',
  )]
  public function transmit(string $report, array $options = ['environment' => 'default']): void {
    $logger = $this->logger();

    if ($this->reporter->transmitReport($report, $options['environment'] ?? 'default')) {
      $logger->notice(sprintf('Report "%s" has been successfully transmitted.', $report));
    }
    else {
      $logger->error('Failed to transmit report data.');
    }
  }

}
