<?php

namespace Drupal\server_beacon\Entity\Form;

use Drupal\Component\Plugin\ConfigurableInterface;
use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\PluginFormFactoryInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Plugin\PluginWithFormsInterface;
use Drupal\server_beacon\Plugin\BeaconTransmitterManagerInterface;
use Drupal\server_beacon\Plugin\ServerBeaconManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Server beacon report entity edit form.
 */
class BeaconReportEdit extends EntityForm {

  /**
   * Creates a new BeaconReportEdit entity form for editing reports.
   *
   * @param \Drupal\Core\Plugin\PluginFormFactoryInterface $pluginFormFactory
   *   The plugin form factory.
   * @param \Drupal\server_beacon\Plugin\BeaconTransmitterManagerInterface $transmitterManager
   *   The beacon transmitter plugins manager service.
   * @param \Drupal\server_beacon\Plugin\ServerBeaconManagerInterface $beaconManager
   *   The server beacon plugin manager service.
   */
  public function __construct(protected PluginFormFactoryInterface $pluginFormFactory, protected BeaconTransmitterManagerInterface $transmitterManager, protected ServerBeaconManagerInterface $beaconManager) {
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): static {
    return new static(
      $container->get('plugin_form.factory'),
      $container->get('plugin.manager.server_beacon_transmitter'),
      $container->get('plugin.manager.server_beacon')
    );
  }

  /**
   * Check if a report name is valid, and available.
   *
   * @param string $id
   *   The report machine name to check for availability.
   *
   * @return bool
   *   TRUE if the machine name can be used, FALSE otherwise.
   */
  public function exists(string $id): bool {
    if ($id === 'add') {
      return TRUE;
    }

    $entityType = $this->getEntity()->getEntityType();
    $entityIds = $this->entityTypeManager
      ->getStorage($entityType->id())
      ->getQuery()
      ->accessCheck(FALSE)
      ->condition($entityType->id(), $id)
      ->range(0, 1)
      ->execute();

    return (bool) $entityIds;
  }

  /**
   * {@inheritdoc}
   */
  protected function copyFormValuesToEntity(EntityInterface $entity, array $form, FormStateInterface $form_state) {
    /** @var \Drupal\server_beacon\Entity\ServerBeaconReportInterface $entity */
    if ($form_state->isProcessingInput()) {
      $values = $form_state->getValues();
      $entityType = $entity->getEntityType();

      // Check and remove the psuedo value.
      if (empty($values['stationUri']['use_path'])) {
        unset($values['stationUri']['path']);
      }
      unset($values['stationUri']['use_path']);

      foreach ($entityType->get('config_export') as $key) {
        if (isset($values[$key])) {
          $entity->set($key, $values[$key]);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state): array {
    $form = parent::form($form, $form_state);
    /** @var \Drupal\server_beacon\Entity\ServerBeaconReportInterface $report */
    $report = $this->buildEntity($form, $form_state);

    $form['name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Report name'),
      '#required' => TRUE,
      '#default_value' => $report->getName(),
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#required' => TRUE,
      '#default_value' => $report->id(),
      '#disabled' => !$report->isNew(),
      '#machine_name' => [
        'source' => ['name'],
        'exists' => [$this, 'exists'],
      ],
    ];

    $stationUri = $report->get('stationUri') ?: [];
    $form['stationUri'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Report to station address'),
      '#tree' => TRUE,

      'uri' => [
        '#type' => 'container',
        '#attributes' => [
          'class' => ['container-inline'],
        ],

        'scheme' => [
          '#type' => 'select',
          '#title' => $this->t('Protocol scheme'),
          '#title_display' => 'invisible',
          '#required' => TRUE,
          '#options' => [
            'http' => 'HTTP',
            'https' => 'HTTPS',
          ],
          '#default_value' => $stationUri['scheme'] ?? 'https',
          '#parents' => ['stationUri', 'scheme'],
        ],
        'host' => [
          '#type' => 'textfield',
          '#title' => $this->t('Target host'),
          '#title_display' => 'invisible',
          '#required' => TRUE,
          '#field_prefix' => '://',
          '#default_value' => $stationUri['host'] ?? NULL,
          '#parents' => ['stationUri', 'host'],
        ],
      ],
      'use_path' => [
        '#id' => 'station-uri-use-path',
        '#type' => 'checkbox',
        '#title' => $this->t('Use a custom path to send this report'),
        '#default_value' => !empty($stationUri['path']),
        '#description' => $this->t('This allows a custom path to be provided for this report. Leaving this empty will default to using the standard Report Station URL path.'),
      ],
      'path' => [
        '#type' => 'textfield',
        '#title' => $this->t('Transmit custom path'),
        '#default_value' => $stationUri['path'] ?? NULL,
        '#states' => [
          'visible' => [
            ':input#station-uri-use-path' => ['checked' => TRUE],
          ],
        ],
      ],
    ];

    $form['transmitterDef'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Report transmitter'),
      '#tree' => TRUE,

      'id' => [
        '#type' => 'select',
        '#title' => $this->t('Protocol'),
        '#required' => TRUE,
        '#options' => $this->transmitterManager->getTransmitterOptions(),
        '#empty_option' => $this->t('- Select Transmitter -'),
      ],
      'config' => [
        '#parents' => ['transmitterDef', 'config'],
        '#array_parents' => ['transmitterDef', 'config'],
      ],
    ];
    $this->buildPluginSubform('report-transmitter-config', $report->getTransmitter(), $form['transmitterDef'], $form, $form_state);

    $form['beaconDefs'] = [];
    if ($beacons = $report->getBeacons()) {
      $form['beaconDefs']['#type'] = 'fieldset';
      $form['beaconDefs']['#title'] = $this->t('Server beacons');
      $form['beaconDefs']['#tree'] = TRUE;
      $form['beaconDefs']['#prefix'] = '<div id="report-beacon-definitions-wrapper">';
      $form['beaconDefs']['#suffix'] = '</div>';

      foreach ($beacons as $bid => $beacon) {
        $form['beaconDefs'][$bid] = [
          '#type' => 'fieldset',
          '#title' => $beacon->getPluginDefinition()['label'] ?? $bid,
          '#parents' => ['beaconDefs', $bid],
          '#array_parents' => ['beaconDefs', $bid],

          'actions' => [
            '#type' => 'actions',

            'remove' => [
              '#type' => 'submit',
              '#name' => "remove-beacon[{$bid}]",
              '#value' => $this->t('Remove Beacon'),
              '#validate' => [[$this, 'removeBeaconValidate']],
              '#limit_validation_errors' => [['beaconDefs']],
              '#submit' => [],
              '#ajax' => [
                'callback' => static::class . '::ajaxAddBeacon',
              ],
            ],
          ],
        ];

        $this->buildPluginConfigForm($beacon, $form['beaconDefs'][$bid], $form, $form_state);
      }
    }
    else {
      $form['beaconDefs'] = [
        '#type' => 'container',
        '#id' => 'report-beacon-definitions-wrapper',
        '#value' => [],
      ];
    }

    // If there are more server beacon types available, provide option to add.
    if ($beaconOpts = array_diff_key($this->beaconManager->getBeaconOptions(), $beacons)) {
      $form['__add_beacon'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Add reporting beacon'),
        '#prefix' => '<div id="report-add-beacon-wrapper">',
        '#suffix' => '</div>',
        '#tree' => TRUE,

        'beacon' => [
          '#theme_wrappers' => [],
          '#type' => 'select',
          '#title' => $this->t('Add beacon'),
          '#title_display' => 'invisible',
          '#options' => $beaconOpts,
        ],
        'add' => [
          '#type' => 'submit',
          '#name' => 'add-beacon',
          '#value' => $this->t('Add beacon'),
          '#validate' => [[$this, 'addBeaconValidate']],
          '#submit' => [],
          '#limit_validation_errors' => [
            ['beaconDefs'],
            ['__add_beacon'],
          ],
          '#ajax' => [
            'callback' => static::class . '::ajaxAddBeacon',
          ],
        ],
      ];
    }
    else {
      $form['__add_beacon'] = [
        '#type' => 'container',
        '#id' => 'report-add-beacon-wrapper',
      ];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    /** @var \Drupal\server_beacon\Entity\ServerBeaconReportInterface $report */
    $report = $this->getEntity();

    // Validate the station domain and path.
    $uri = $form_state->getValue('stationUri');
    if (FALSE === filter_var($uri['host'], FILTER_VALIDATE_DOMAIN)) {
      $form_state->setError($form['stationUri']['host'], $this->t('Invalid host domain specified. Textfield should only container domain value.'));
    }
    if ($uri['use_path'] && FALSE === filter_var($uri['path'], FILTER_VALIDATE_URL)) {
      $form_state->setError($form['stationUri']['path'], $this->t('Invalid API path. Not a valid URL path.'));
    }

    $transmitter = $report->getTransmitter();
    if ($transmitter && ($pluginForm = $this->getPluginConfigurationForm($transmitter))) {
      $subform = &$form['transmitterDef']['config'];
      $subformState = SubformState::createForSubform($subform, $form, $form_state);
      $pluginForm->validateConfigurationForm($subform, $subformState);
    }

    if ($report->getBeacons()) {
      /** @var \Drupal\server_beacon\Plugin\ServerBeaconInterface $beacon */
      foreach ($report->getBeacons() as $bid => $beacon) {
        if ($beaconForm = $this->getPluginConfigurationForm($beacon)) {
          $subform = &$form['beaconDefs'][$bid];
          $subformState = SubformState::createForSubform($subform, $form, $form_state);
          $beaconForm->validateConfigurationForm($subform, $subformState);
        }
      }
    }
    else {
      $form_state->setError($form['__add_beacon'], $this->t('Report should contain at least one report beacon.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state): int {
    /** @var \Drupal\server_beacon\Entity\ServerBeaconReportInterface $report */
    $report = $this->entity;

    $transmitter = $report->getTransmitter();
    if ($pluginForm = $this->getPluginConfigurationForm($transmitter)) {
      $subform = &$form['transmitterDef']['config'];
      $subformState = SubformState::createForSubform($subform, $form, $form_state);
      $pluginForm->submitConfigurationForm($subform, $subformState);

      $report->set('transmitterDef', [
        'id' => $transmitter->getPluginId(),
        'config' => $transmitter->getConfiguration(),
      ]);
    }

    $beaconDefs = [];
    /** @var \Drupal\server_beacon\Plugin\ServerBeaconInterface $beacon */
    foreach ($report->getBeacons() as $bid => $beacon) {
      $beaconForm = $this->getPluginConfigurationForm($beacon);

      if ($beaconForm && ($beacon instanceof ConfigurableInterface)) {
        $subform = &$form['beaconDefs'][$bid];
        $subformState = SubformState::createForSubform($subform, $form, $form_state);
        $beaconForm->submitConfigurationForm($subform, $subformState);
        $beaconDefs[$bid] = $beacon->getConfiguration();
      }
      else {
        $beaconDefs[$bid] = [];
      }
    }
    $report->set('beaconDefs', $beaconDefs);

    $msgParams['%report'] = $report->label();
    $msgParams['@type'] = $report->getEntityType()->getLabel();
    $message = $report->isNew()
      ? $this->t('Successfully created new @type %report.', $msgParams)
      : $this->t('Updates to @type %report were successfully saved.', $msgParams);

    $this->messenger()->addMessage($message);
    $form_state->setRedirectUrl($report->toUrl('collection'));

    return parent::save($form, $form_state);
  }

  /**
   * Form validate callback to handle the addition of beacon plugins.
   *
   * This form callback requests a form rebuild, and will skip any form
   * submit callbacks.
   *
   * @param array $form
   *   The form elements and structure.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state and values.
   */
  public function addBeaconValidate(array &$form, FormStateInterface $form_state): void {
    $beaconId = $form_state->getValue(['__add_beacon', 'beacon']);
    $beaconDefs = $form_state->getValue('beaconDefs');
    $beaconDefs[$beaconId] = [];

    $form_state
      ->setValue('beaconDefs', $beaconDefs)
      ->setRebuild(TRUE);
  }

  /**
   * Form validate callback to handle the addition of beacon plugins.
   *
   * This form callback requests a form rebuild, and will skip any form
   * submit callbacks.
   *
   * @param array $form
   *   The form elements and structure.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state and values.
   */
  public function removeBeaconValidate(array &$form, FormStateInterface $form_state): void {
    $trigger = $form_state->getTriggeringElement();

    if (preg_match('/^remove-beacon\[([^\]]+)\]$/', $trigger['#name'], $matches)) {
      $beaconDefs = $form_state->getValue('beaconDefs');
      unset($beaconDefs[$matches[1]]);

      $form_state
        ->setValue('beaconDefs', $beaconDefs)
        ->setRebuild(TRUE);
    }
  }

  /**
   * Update plugin configuration through form AJAX.
   *
   * @param array $form
   *   The complete for elements array.
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The submitted form state.
   *
   * @return array|\Drupal\Core\Ajax\AjaxResponse
   *   Either a renderable array of elements to replace the config elements with
   *   or a set of AJAX commands wrapped in a response that
   */
  public static function pluginAjaxUpdate(array $form, FormStateInterface $formState): array|AjaxResponse {
    $parents = $formState->getTriggeringElement()['#array_parents'];
    array_pop($parents);
    $parents[] = 'config';

    return NestedArray::getValue($form, $parents);
  }

  /**
   * Form AJAX callback to update the form for beacon addition.
   *
   * @param array $form
   *   The form elements and structure of the report edit form.
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The current form state.
   *
   * @return array|\Drupal\Core\Ajax\AjaxResponse
   *   An AJAX response to replace to add a new beacon plugin handler.
   */
  public static function ajaxAddBeacon(array $form, FormStateInterface $formState): array|AjaxResponse {
    return (new AjaxResponse())
      ->addCommand(new ReplaceCommand('#report-beacon-definitions-wrapper', $form['beaconDefs']))
      ->addCommand(new ReplaceCommand('#report-add-beacon-wrapper', $form['__add_beacon']));
  }

  /**
   * Get the plugin configuration form instance if the plugin implements one.
   *
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The plugin to locate a plugin configuration form instance for.
   *
   * @return \Drupal\Core\Plugin\PluginFormInterface|null
   *   The plugin form instance that builds the plugin configuration if the
   *   plugin provides a configuration form builder. NULL otherwise.
   */
  protected function getPluginConfigurationForm(PluginInspectionInterface $plugin): ?PluginFormInterface {
    if ($plugin instanceof PluginFormInterface) {
      return $plugin;
    }
    if ($plugin instanceof PluginWithFormsInterface) {
      return $this->pluginFormFactory->createInstance($plugin, 'configure');
    }

    return NULL;
  }

  /**
   * Create a plugin form for a report transmitter including plugin selector.
   *
   * @param string $wrapperId
   *   The form element ID which wraps the configuration elements. Used for
   *   the AJAX wrapper identifier.
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The plugin to build the plugin form elements to.
   * @param array $subform
   *   Reference to the form elements to use to build the plugin form. The form
   *   should minimally include the '#array_parents and #parents attributes.
   * @param array $completeForm
   *   The complete form structure and object.
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The complete form state.
   */
  protected function buildPluginSubform(string $wrapperId, ?PluginInspectionInterface $plugin, array &$subform, array $completeForm, FormStateInterface $formState): void {
    if ($plugin) {
      $subform['id']['#default_value'] = $plugin->getPluginId();
      $subform['config'] += [
        '#type' => 'fieldset',
        '#title' => $this->t('Configure'),
        '#title_display' => 'invisible',
      ];
      $this->buildPluginConfigForm($plugin, $subform['config'], $completeForm, $formState);
    }

    $subform['config']['#prefix'] = '<div id="' . $wrapperId . '">';
    $subform['config']['#suffix'] = '</div>';
    $subform['id']['#ajax'] = [
      'wrapper' => $wrapperId,
      'callback' => static::class . '::pluginAjaxUpdate',
    ];
  }

  /**
   * Create a plugin configuration form for a report transmitter or beacon.
   *
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The plugin to build and apply configuration form elements to.
   * @param array $configForm
   *   Reference to the form elements to use to build the plugin configuration
   *   form. The form should minimally include the '#array_parents and #parents.
   * @param array $completeForm
   *   The complete form structure and object.
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The complete form state.
   */
  protected function buildPluginConfigForm(PluginInspectionInterface $plugin, array &$configForm, array $completeForm, FormStateInterface $formState): void {
    if ($pluginForm = $this->getPluginConfigurationForm($plugin)) {
      $configFormState = SubformState::createForSubform($configForm, $completeForm, $formState);
      $configForm = $pluginForm->buildConfigurationForm($configForm, $configFormState) + [
        '#type' => 'container',
        '#attributes' => [],
        '#description' => $plugin->getPluginDefinition()['description'] ?? NULL,
        '#description_display' => 'before',
      ];
    }
    else {
      $configForm['#markup'] = $this->t('Server beacon does not require additional settings.');
      $configForm['#value'] = [];
    }
  }

}
