<?php

namespace Drupal\server_beacon\Entity;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Entity\Attribute\ConfigEntityType;
use Drupal\Core\Entity\Routing\AdminHtmlRouteProvider;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\server_beacon\Entity\Controller\BeaconReportListBuilder;
use Drupal\server_beacon\Entity\Form\BeaconReportDeleteConfirm;
use Drupal\server_beacon\Entity\Form\BeaconReportEdit;
use Drupal\server_beacon\Plugin\BeaconTransmitterInterface;
use Drupal\server_beacon\Plugin\BeaconTransmitterManagerInterface;
use Drupal\server_beacon\Plugin\ServerBeaconManagerInterface;

/**
 * Configuration entity to define a server beacon report.
 *
 * The report include server beacons to be constructed and generating report
 * data as well as the target server station that is expected to recieve the
 * beacon transmissions.
 *
 * @ConfigEntityType(
 *   id = "server_beacon_report",
 *   label = @Translation("Server report"),
 *   config_prefix = "report",
 *   admin_permission = "administer configurations",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "name",
 *   },
 *   handlers = {
 *    "list_builder" = "Drupal\server_beacon\Entity\Controller\BeaconReportListBuilder",
 *    "form" = {
 *      "default" = "Drupal\server_beacon\Entity\Form\BeaconReportEdit",
 *      "edit" = "Drupal\server_beacon\Entity\Form\BeaconReportEdit",
 *      "delete" = "Drupal\server_beacon\Entity\Form\BeaconReportDeleteConfirm",
 *    },
 *    "route_provider" = {
 *      "html" = "Drupal\Core\Entity\Routing\AdminHtmlRouteProvider",
 *    },
 *   },
 *   links = {
 *     "collection" = "/admin/config/services/server-beacon",
 *     "add-form" = "/admin/config/services/server-beacon/add",
 *     "edit-form" = "/admin/config/services/server-beacon/{server_beacon_report}",
 *     "delete-form" = "/admin/config/services/server-beacon/{server_beacon_report}/delete",
 *   },
 *   config_export = {
 *     "id",
 *     "name",
 *     "stationUri",
 *     "transmitterDef",
 *     "beaconDefs",
 *   },
 * )
 */
#[ConfigEntityType(
  id: 'server_beacon_report',
  label: new TranslatableMarkup('Server report'),
  config_prefix: 'report',
  admin_permission: 'administer configurations',
  entity_keys: [
    'id' => 'id',
    'label' => 'name',
  ],
  handlers: [
    'list_builder' => BeaconReportListBuilder::class,
    'form' => [
      'default' => BeaconReportEdit::class,
      'edit' => BeaconReportEdit::class,
      'delete' => BeaconReportDeleteConfirm::class,
    ],
    'route_provider' => [
      'html' => AdminHtmlRouteProvider::class,
    ],
  ],
  links: [
    'collection' => '/admin/config/services/server-beacon',
    'add-form' => '/admin/config/services/server-beacon/add',
    'edit-form' => '/admin/config/services/server-beacon/{server_beacon_report}',
    'delete-form' => '/admin/config/services/server-beacon/{server_beacon_report}/delete',
  ],
  config_export: [
    'id',
    'name',
    'stationUri',
    'transmitterDef',
    'beaconDefs',
  ],
)]
class ServerBeaconReport extends ConfigEntityBase implements ServerBeaconReportInterface {

  use StringTranslationTrait;
  use LoggerChannelTrait;

  /**
   * The server beacon report identifier.
   *
   * @var string
   */
  protected string $id;

  /**
   * Friendly display label for the beacon report.
   *
   * @var \Stringable|string
   */
  protected \Stringable|string $name = '';

  /**
   * Information about the server station to send the report to.
   *
   * @var mixed[]
   */
  protected array $stationUri;

  /**
   * The beacon transmitter plugin definition.
   */
  protected array $transmitterDef = [];

  /**
   * The loaded and configure beacon transmitter to use for sending reports.
   *
   * @var \Drupal\server_beacon\Plugin\BeaconTransmitterInterface
   */
  protected ?BeaconTransmitterInterface $transmitter = NULL;

  /**
   * The beacon definitions for config storage and loading.
   *
   * @var mixed[]
   */
  protected array $beaconDefs = [];

  /**
   * Load server beacon handler instances.
   *
   * @var \Drupal\server_beacon\Plugin\ServerBeaconInterface[]|null
   */
  protected ?array $beacons = NULL;

  /**
   * Get the server beacon manager instance.
   *
   * @return \Drupal\server_beacon\Plugin\ServerBeaconManagerInterface
   *   Get the server beacon plugin manager.
   */
  protected static function getBeaconManager(): ServerBeaconManagerInterface {
    /** @var \Drupal\server_beacon\Plugin\ServerBeaconManagerInterface $beaconManager */
    static $beaconManager = \Drupal::service('plugin.manager.server_beacon');
    return $beaconManager;
  }

  /**
   * Get the server beacon manager instance.
   *
   * @return \Drupal\server_beacon\Plugin\BeaconTransmitterManagerInterface
   *   Get the server beacon transmitter plugin manager.
   */
  protected static function getTransmitterManager(): BeaconTransmitterManagerInterface {
    /** @var \Drupal\server_beacon\Plugin\BeaconTransmitterManagerInterface $beaconManager */
    static $beaconManager = \Drupal::service('plugin.manager.server_beacon_transmitter');
    return $beaconManager;
  }

  /**
   * {@inheritdoc}
   */
  public function getName(): string {
    return $this->name;
  }

  /**
   * {@inheritdoc}
   */
  public function reportHost(): string {
    @['scheme' => $scheme, 'host' => $host] = $this->stationUri;
    return "{$scheme}://{$host}";
  }

  /**
   * {@inheritdoc}
   */
  public function reportTo(): string {
    ['scheme' => $scheme, 'host' => $host, 'path' => $path] = $this->stationUri + [
      'path' => NULL,
    ];
    $url = "{$scheme}://{$host}";
    $url .= empty($path) ? '/api/report-station/' . $this->id() : $path;

    return $url;
  }

  /**
   * {@inheritdoc}
   */
  public function getBeaconTypes(bool $label = FALSE): array {
    $beaconManager = static::getBeaconManager();
    $propertyName = $label ? 'label' : 'id';

    $types = [];
    foreach ($this->beaconDefs as $id => $def) {
      $beaconDef = $beaconManager->getDefinition($id, FALSE);
      $types[$id] = $beaconDef[$propertyName];
    }

    return $types;
  }

  /**
   * {@inheritdoc}
   */
  public function set($property_name, $value): self {
    parent::set($property_name, $value);

    // Reset the server beacons attached to this report.
    switch ($property_name) {
      case 'beaconDefs':
        $this->beacons = NULL;
        break;

      case 'transmitterDef':
        $this->transmitter = NULL;
        break;
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getTransmitter(): ?BeaconTransmitterInterface {
    if (empty($this->transmitterDef['id'])) {
      return NULL;
    }

    if (!isset($this->transmitter)) {
      ['id' => $id, 'config' => $config] = $this->transmitterDef + [
        'config' => [],
      ];
      $this->transmitter = $this->getTransmitterManager()->createInstance($id, $config);
    }

    return $this->transmitter;
  }

  /**
   * {@inheritdoc}
   */
  public function getBeacons(): array {
    if (!isset($this->beacons)) {
      $beaconManager = static::getBeaconManager();

      $this->beacons = [];
      foreach ($this->beaconDefs as $pluginId => $config) {
        try {
          $this->beacons[$pluginId] = $beaconManager->createInstance($pluginId, $config ?? []);
        }
        catch (PluginNotFoundException $e) {
          // Skip invalid or missing beacon plugin types.
        }
      }
    }

    return $this->beacons;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies(): self {
    parent::calculateDependencies();

    if ($transmitter = $this->getTransmitter()) {
      $this->calculatePluginDependencies($transmitter);
    }

    foreach ($this->getBeacons() as $beacon) {
      $this->calculatePluginDependencies($beacon);
    }

    return $this;
  }

}
