<?php

namespace Drupal\server_beacon\Plugin;

use Drupal\Component\Plugin\ConfigurableInterface;
use Drupal\Component\Plugin\Definition\PluginDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Plugin\PluginFormInterface;

/**
 * A base plugin for Server Beacon plugins which are configurable.
 *
 * Provides the most common functionality so each plugin doesn't have to
 * redefine a default implementation.
 */
abstract class ConfigurablePluginBase extends PluginBase implements ConfigurableInterface, PluginFormInterface {

  /**
   * Creates a new instance of the configurable plugin object.
   *
   * @param array $configuration
   *   The plugin configuration.
   * @param string $plugin_id
   *   The plugin unique identifier.
   * @param array|\Drupal\Component\Plugin\Definition\PluginDefinitionInterface $plugin_definition
   *   The plugin definition.
   */
  public function __construct(array $configuration, string $plugin_id, array|PluginDefinitionInterface $plugin_definition) {
    parent::__construct(
      $configuration + $this->defaultConfiguration(),
      $plugin_id,
      $plugin_definition
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration(): array {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration): void {
    $this->configuration = $configuration + $this->defaultConfiguration();
  }

  /**
   * Process and restructure plugin configuration form values for plugin config.
   *
   * @param array $values
   *   Raw configuration values to process.
   * @param \Drupal\Core\Form\FormStateInterface|null $formState
   *   The form state for the submitted configuration form (could be subform).
   *
   * @return array
   *   The processed configuration values ready for applying to the plugin.
   */
  public function massageConfigurationValues(array $values, ?FormStateInterface $formState = NULL): array {
    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state): void {
    // Most plugin configuration validation is handled by the FAPI elements.
  }

  /**
   * {@inheritdoc}
   *
   * Use the static::defaultConfiguration() keys to determine which form values
   * to transfer to the plugin configuration.
   *
   * Use the static::massageConfigurationValues() in order restructure the form
   * submitted values into the expected format for the plugin configuration.
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $this->massageConfigurationValues($form_state->getValues(), $form_state);
    $values = array_intersect_key($values, $this->defaultConfiguration());

    $this->setConfiguration($values);
  }

}
