<?php

namespace Drupal\server_beacon\Plugin\KeyType;

use Drupal\Core\Entity\EntityFormInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\key\Plugin\KeyPluginFormInterface;
use Drupal\key\Plugin\KeyTypeBase;

/**
 * Defines a generic key type for encryption.
 *
 * @KeyType(
 *   id = "server_beacon_signkey",
 *   label = @Translation("Generated signing keypair"),
 *   description = @Translation("Generates a keypair to signing and verifying report transmissions."),
 *   group = "encryption",
 *   key_value = {
 *     "plugin" = "generate",
 *   }
 * )
 */
class BeaconSignKey extends KeyTypeBase implements KeyPluginFormInterface {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'type' => 'ed25519',
    ];
  }

  /**
   * Get the list of key type options available for generated key type.
   *
   * @return array<string,\Stringable|string>
   *   List of values compatible with form API options.
   */
  protected function getTypeOptions(): array {
    return [
      'ed25519' => $this->t('@key_type keypair', ['@key_type' => 'ed25519']),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $formObj = $form_state->getFormObject();
    /** @var \Drupal\key\Entity\Key $key */
    $key = $formObj instanceof EntityFormInterface ? $formObj->getEntity() : NULL;

    $form['type'] = [
      '#type' => 'select',
      '#title' => $this->t('Key type'),
      '#required' => TRUE,
      '#options' => $this->getTypeOptions(),
      '#default_value' => $this->configuration['type'] ?? 'ed25519',
    ];

    if ($key && ($keypair = $key->getKeyValue())) {
      $form['type']['#disabled'] = TRUE;
      $form['type']['#description'] = $this->t('Keypair has been generated, so the key type can no longer be changed.');

      $publicKey = \sodium_crypto_sign_publickey($keypair);
      $form['public_key_display'] = [
        '#theme' => 'server_beacon_public_key',
        '#public_key' => \sodium_bin2base64($publicKey, SODIUM_BASE64_VARIANT_ORIGINAL),
      ];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state): void {
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->setConfiguration($form_state->getValues());
  }

  /**
   * {@inheritdoc}
   */
  public function validateKeyValue(array $form, FormStateInterface $form_state, $key_value): void {
  }

  /**
   * {@inheritdoc}
   */
  public static function generateKeyValue(array $configuration): string {
    return \sodium_crypto_sign_keypair();
  }

}
