<?php

namespace Drupal\server_beacon\Plugin\ServerBeacon\Beacon;

use Drupal\Component\Plugin\Definition\PluginDefinitionInterface;
use Drupal\Component\Serialization\SerializationInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\server_beacon\Attribute\ServerBeacon;
use Drupal\server_beacon\Plugin\ConfigurablePluginBase;
use Drupal\server_beacon\Plugin\ConfigurableServerBeaconInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Tinkersmith\Console\Cli;
use Tinkersmith\Console\Exception\CliRuntimeException;
use Tinkersmith\Console\Output\BufferedOutput;

/**
 * Server beacon for generating and displaying Composer Audit information.
 */
#[ServerBeacon(
  id: 'composer',
  label: new TranslatableMarkup('Composer Audit'),
  description: new TranslatableMarkup('Provides information form Composer Audits in JSON format.'),
)]
class ComposerServerBeacon extends ConfigurablePluginBase implements ConfigurableServerBeaconInterface, ContainerFactoryPluginInterface {

  const MAX_COMPOSER_PARENTS = 2;

  /**
   * Creates a new instance of the ComposerServerBeacon plugin handler.
   *
   * @param array $configuration
   *   The server beacon pluign configurations.
   * @param string $pluginId
   *   The server beacon plugin identifier.
   * @param array|\Drupal\Component\Plugin\Definition\PluginDefinitionInterface $definition
   *   The server beacon definition.
   * @param string $appRoot
   *   The path to the Drupal application root directory.
   * @param \Drupal\Component\Serialization\SerializationInterface $json
   *   JSON supported serialization service.
   */
  public function __construct(
    array $configuration,
    string $pluginId,
    array|PluginDefinitionInterface $definition,
    protected readonly string $appRoot,
    protected SerializationInterface $json,
  ) {
    $configuration += $this->defaultConfiguration();
    parent::__construct($configuration, $pluginId, $definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $pluginId, $definition): static {
    return new static(
      $configuration,
      $pluginId,
      $definition,
      $container->getParameter('app.root'),
      $container->get('serialization.json')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'securityAudit' => TRUE,
      'outdated' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function fetchStatus(?OutputInterface $output = NULL): array {
    $config = $this->getConfiguration();
    $cli = new Cli(['composer'], $this->findComposer());

    $data = [];
    if (!empty($config['securityAudit'])) {
      try {
        $buffer = new BufferedOutput($output);
        $cli->execute(['audit', '--no-dev', '--no-scripts', '--format=json'], $buffer);
        $data = $this->json->decode($buffer->fetch()) ?: [];
      }
      catch (CliRuntimeException $e) {
        $process = $e->getProcess();
        if (!$process->isOutputDisabled()) {
          $data = $this->json->decode($process->getOutput()) ?: [];
        }

        // If the error was not the composer audit flags, then this is an
        // unexpected error. That needs to be sent to the command.
        if (!(isset($data['advisories']) || isset($data['abandoned']))) {
          throw $e;
        }
      }
    }
    if (!empty($config['outdated'])) {
      $args = ['outdated', '--no-dev', '--no-scripts', '--format=json'];
      if ('direct' === $config['outdated']) {
        $args[] = '--direct';
      }

      $buffer = new BufferedOutput($output);
      $cli->execute($args, $buffer);
      $data['outdated'] = [
        'showing' => $config['outdated'],
        'packages' => $this->json->decode($buffer->fetch())['installed'] ?? [],
      ];
    }

    return array_filter($data);
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $config = $this->getConfiguration();

    $form['securityAudit'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include security audit (recommended)'),
      '#default_value' => $config['securityAudit'],
    ];
    $form['outdated'] = [
      '#type' => 'select',
      '#title' => $this->t('Include composer outdated packages'),
      '#options' => [
        '' => $this->t('None'),
        'direct' => $this->t('Only direct composer.json dependencies'),
        'all' => $this->t('All outdated dependencies'),
      ],
      '#default_value' => $config['outdated'],
    ];

    return $form;
  }

  /**
   * Find the base composer directory as a parent of the $appRoot.
   *
   * Searches through parent directories for a "composer.json" file. Will use
   * the first parent directory or current app directory that has a
   * "composer.json" file. Limited searching only static::MAX_COMPOSER_PARENTS
   * parent directories.
   *
   * @return string
   *   Provide the working directory for Composer commands.
   *
   * @throws \InvalidArgumentException
   *   When the Composer path cannot be determined automatically.
   */
  protected function findComposer(): string {
    $composerRoot = realpath($this->appRoot);

    $i = 0;
    while (TRUE) {
      // We found the composer directory for the project.
      if (file_exists($composerRoot . '/composer.json')) {
        return $composerRoot;
      }

      $composerRoot = dirname($composerRoot);
      if (!strlen($composerRoot) && ++$i > static::MAX_COMPOSER_PARENTS) {
        throw new \InvalidArgumentException('Unable to find Composer directory.');
      }
    }
  }

}
