<?php

namespace Drupal\server_beacon\Plugin\ServerBeacon\Beacon;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Plugin\Definition\PluginDefinitionInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Install\Tasks;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\server_beacon\Attribute\ServerBeacon;
use Drupal\server_beacon\Plugin\ConfigurablePluginBase;
use Drupal\server_beacon\Plugin\ConfigurableServerBeaconInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Server beacon for generating and displaying Drupal and contrib module status.
 */
#[ServerBeacon(
  id: 'drupal',
  label: new TranslatableMarkup('Drupal status'),
  description: new TranslatableMarkup('Provides information about Drupal core and contributed module status.'),
)]
class DrupalServerBeacon extends ConfigurablePluginBase implements ConfigurableServerBeaconInterface, ContainerFactoryPluginInterface {

  /**
   * Creates a new instance of the PhpServerBeacon plugin handler.
   *
   * @param array $configuration
   *   The server beacon pluign configurations.
   * @param string $pluginId
   *   The server beacon plugin identifier.
   * @param array|\Drupal\Component\Plugin\Definition\PluginDefinitionInterface $definition
   *   The server beacon definition.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The path to the Drupal application root directory.
   * @param \Drupal\Core\Database\Connection $db
   *   The default Drupal database connection.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state storage interface.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The system date and time helper service.
   */
  public function __construct(
    array $configuration,
    string $pluginId,
    array|PluginDefinitionInterface $definition,
    protected ModuleHandlerInterface $moduleHandler,
    protected Connection $db,
    protected StateInterface $state,
    protected TimeInterface $time,
  ) {
    parent::__construct($configuration, $pluginId, $definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $pluginId, $definition): static {
    return new static(
      $configuration,
      $pluginId,
      $definition,
      $container->get('module_handler'),
      $container->get('database'),
      $container->get('state'),
      $container->get('datetime.time')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'db_info' => TRUE,
      'contrib' => FALSE,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function fetchStatus(): array {
    $config = $this->getConfiguration();
    $status = ['drupal' => \Drupal::VERSION];

    if (!empty($config['db_info'])) {
      $class = $this->db->getConnectionOptions()['namespace'] . '\\Install\\Tasks';
      $status['db'] = [
        'driver' => is_a($class, Tasks::class, TRUE) ? (new $class())->name() : 'Database',
        'version' => $this->db->version(),
      ];

      if (preg_match('/[\d.]+/', $status['db']['version'], $matches)) {
        $status['db']['version'] = $matches[0];
      }
    }

    if (!empty($config['contrib']) && $this->moduleHandler->moduleExists('update')) {
      $this->moduleHandler->loadInclude('update', 'compare.inc');
      $lastCheck = $this->state->get('update.last_check') ?: 0;
      $currTime = $this->time->getRequestTime();

      // Refresh the available modules if last check was more than a day ago.
      if ($available = update_get_available(($currTime - $lastCheck) > 86400)) {
        $status['contrib'] = [];
        foreach (update_calculate_project_data($available) as $module => $info) {
          if (0 <= $info['status'] && $info['existing_version'] !== $info['latest_version']) {
            $release = $available[$module]['releases'][$info['existing_version']] ?? [];
            $status['contrib'][$module] = [
              'name' => $info['title'] ?? $module,
              'type' => $info['project_type'],
              'version' => $info['existing_version'],
              'latest' => $info['latest_version'],
              'insecure' => in_array('Insecure', $release['terms']['Release type'] ?? []),
            ];
          }
        }
      }
      else {
        $status['_msg']['contrib'] = 'Unable to load or fetch available update information.';
      }
    }

    return $status;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $config = $this->getConfiguration();

    $form['db_info'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include database version and type'),
      '#default_value' => $config['db_info'],
    ];
    $form['contrib'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include security & upgrade data for installed contributed modules'),
      '#default_value' => $config['contrib'],
      '#description' => $this->t('When available, PHP Composer audits are preferred since it provides information about all dependencies and outdated packages. This option requires "Update manager" module to be enabled.'),
    ];

    return $form;
  }

}
