<?php

namespace Drupal\server_beacon\Plugin\ServerBeacon\Beacon;

use Drupal\Component\Plugin\Definition\PluginDefinitionInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\server_beacon\Attribute\ServerBeacon;
use Drupal\server_beacon\Plugin\ConfigurablePluginBase;
use Drupal\server_beacon\Plugin\ConfigurableServerBeaconInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Dotenv\Dotenv;
use Tinkersmith\Console\Cli;
use Tinkersmith\Console\Exception\CliRuntimeException;
use Tinkersmith\Console\Output\BufferedOutput;

/**
 * Server beacon for generating and displaying PHP and webserver information.
 */
#[ServerBeacon(
  id: 'php',
  label: new TranslatableMarkup('PHP Info'),
  description: new TranslatableMarkup('Provides info about PHP versions and optionally the webserver.'),
)]
class PhpServerBeacon extends ConfigurablePluginBase implements ConfigurableServerBeaconInterface, ContainerFactoryPluginInterface {

  const OS_RELEASE_FILE = '/etc/os-release';

  /**
   * Creates a new instance of the PhpServerBeacon plugin handler.
   *
   * @param array $configuration
   *   The server beacon pluign configurations.
   * @param string $pluginId
   *   The server beacon plugin identifier.
   * @param array|\Drupal\Component\Plugin\Definition\PluginDefinitionInterface $definition
   *   The server beacon definition.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The path to the Drupal application root directory.
   */
  public function __construct(array $configuration, string $pluginId, array|PluginDefinitionInterface $definition, protected ModuleHandlerInterface $moduleHandler) {
    parent::__construct($configuration, $pluginId, $definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $pluginId, $definition): static {
    return new static(
      $configuration,
      $pluginId,
      $definition,
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'osInfo' => FALSE,
      'webserver' => '',
      'includeExtensions' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function fetchStatus(): array {
    $status = [];
    $config = $this->getConfiguration();

    // Get list of PHP extensions to fetch versions for.
    $extensions = $config['includeExtensions'] ?? [];

    // Get versions for listed PHP extensions and if they are available.
    $status['extensions']['php'] = phpversion();
    foreach ($extensions as $extName) {
      $status['extensions'][$extName] = phpversion($extName) ?: 'not installed';
    }

    // If possibly request webserver information.
    try {
      $buffer = new BufferedOutput();
      if ('nginx' === $config['webserver']) {
        (new Cli('nginx'))->execute(['-v'], $buffer);
      }
      elseif ('httpd' === $config['webserver']) {
        try {
          (new Cli('httpd'))->execute(['-v'], $buffer);
        }
        catch (CliRuntimeException $e) {
          (new Cli('apache2'))->execute(['-v'], $buffer);
        }
      }
      // Pass the server info if there was a successful CLI call.
      [$webserver, $serverBuild] = array_map(
        fn($val) => explode(':', $val, 2),
        preg_split("/\r?\n/i", $buffer->fetch(), -1, PREG_SPLIT_NO_EMPTY),
      );
      $status['webserver']['type'] = trim(array_pop($webserver) ?? '');
      $status['webserver']['build'] = trim(array_pop($serverBuild) ?? '');
    }
    catch (CliRuntimeException $e) {
    }

    // Attempt to discover OS information if requested.
    if (!empty($config['osInfo'])) {
      try {
        if (is_readable(self::OS_RELEASE_FILE)) {
          $content = file_get_contents(self::OS_RELEASE_FILE);
          $values = (new Dotenv())->parse($content, self::OS_RELEASE_FILE);
          $status['os'] = array_intersect_key($values, [
            'ID' => TRUE,
            'ID_LIKE' => TRUE,
            'VERSION_ID' => TRUE,
            'PRETTY_NAME' => TRUE,
          ]);
        }
        else {
          $map = [
            'Distributor ID' => 'ID',
            'Release' => 'VERSION_ID',
            'Description' => 'PRETTY_NAME',
          ];

          $buffer = new BufferedOutput();
          (new Cli('lsb_release'))->execute(['-dri'], $buffer);

          // Capture each available line of the OS data. Keep a minimal
          // set of data, but enough to know if things need to be checked.
          foreach (explode("\n", $buffer->fetch()) as $row) {
            [$propName, $value] = explode(':', $row, 2);
            $key = $map[$propName] ?? $propName;
            $status['os'][$key] = trim($value);
          }
        }
      }
      catch (CliRuntimeException $e) {
        $stats['os']['ID'] = 'Unavailable';
      }
    }

    return $status;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $config = $this->getConfiguration();

    $form['webserver'] = [
      '#type' => 'select',
      '#title' => $this->t('Include basic webserver info'),
      '#options' => [
        '' => $this->t('None'),
        'nginx' => $this->t('NGinX'),
        'httpd' => $this->t('Apache HTTPD'),
      ],
      '#default_value' => $config['webserver'] ?? '',
    ];

    $form['includeExtensions'] = [
      '#type' => 'textarea',
      '#title' => $this->t('PHP Extensions'),
      '#default_value' => is_array($config['includeExtensions'])
        ? implode("\n", $config['includeExtensions'] ?? [])
        : $config['includeExtensions'],
      '#description' => $this->t('Enter the names of PHP extensions that should be included with the report, one extension name per line.'),
    ];

    $form['osInfo'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include Linux OS info'),
      '#default_value' => $config['osInfo'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $defaults = $this->defaultConfiguration();

    $config = [];
    foreach (array_keys($defaults) as $key) {
      $config[$key] = $form_state->getValue($key);
    }
    $config['includeExtensions'] = array_map(trim(...), explode("\n", $config['includeExtensions']));
    $config['includeExtensions'] = array_filter($config['includeExtensions']);

    $this->setConfiguration($config + $defaults);
  }

}
