<?php

namespace Drupal\server_beacon\Plugin\ServerBeacon\Transmitter;

use Drupal\Component\Plugin\Definition\PluginDefinitionInterface;
use Drupal\Component\Serialization\SerializationInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\key\KeyRepositoryInterface;
use Drupal\server_beacon\Attribute\ServerBeaconTransmitter;
use Drupal\server_beacon\Entity\ServerBeaconReportInterface;
use Drupal\server_beacon\Plugin\BeaconTransmitterInterface;
use Drupal\server_beacon\Plugin\ConfigurablePluginBase;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\RequestOptions;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Asymeterical encrypted transmission using libsodium.
 *
 * First version of the asymetrical server beacon transmitter using the
 * libsodium library. The request will encrypt using a server station public
 * key and signs with the local private key.
 *
 * The receiving station will decrypt with the station's private key, and verify
 * the message signature with beacon's public key.
 *
 * Sodium also supports an synchronous encryption. Use the "salty_1" (Salty1)
 * transmitter to use the synchronous encryption (same keys on both ends).
 */
#[ServerBeaconTransmitter(
  id: 'asalty_v1',
  label: new TranslatableMarkup('Public key w/ signature (v1.0)'),
  description: new TranslatableMarkup('Uses public and private keys to sign and encrypt the report transmissions. Signing keys are created and owned by the reporter, and only requires the public key of the report receiver.'),
)]
class ASalty1 extends ConfigurablePluginBase implements BeaconTransmitterInterface, ContainerFactoryPluginInterface {

  /**
   * Create a new instance of the Salty v1 transmitter.
   *
   * @param array $configuration
   *   The configuration settings.
   * @param string $plugin_id
   *   The transmitter plugin identifier.
   * @param array|\Drupal\Component\Plugin\Definition\PluginDefinitionInterface $plugin_definition
   *   The plugin definition.
   * @param \Drupal\Component\Serialization\SerializationInterface $serializer
   *   The data serializer for encoding the data to a encryptable string.
   * @param \Drupal\key\KeyRepositoryInterface $keyRepository
   *   The encryption and signing key manager.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    array|PluginDefinitionInterface $plugin_definition,
    protected SerializationInterface $serializer,
    protected KeyRepositoryInterface $keyRepository,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('serialization.json'),
      $container->get('key.repository')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'signingKey' => NULL,
      'stationPublicKey' => NULL,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function __invoke(ServerBeaconReportInterface $report, ClientInterface $client, array $data): bool {
    $signKey = $this->getSigningKey();
    $encryptKey = $this->getPublicEncryptKey();

    if ($signKey && $encryptKey) {
      // Encode the report data into JSON format and then sign it with the
      // private signing key. This ensures that the message was created with
      // by this server and has not been tampered with (assuming private key not
      // leaked or cracked).
      $encoded = $this->serializer->encode($data);
      $signed = \sodium_crypto_sign($encoded, $signKey);

      // Transmit encrypted status report to the service station.
      $response = $client->request('POST', $report->reportTo(), [
        RequestOptions::HEADERS => [
          'Content-Type' => 'application/octet-stream',
        ],
        // Encrypt with public report station key, only the private station key
        // should be able to decrypt and read the message.
        RequestOptions::BODY => \sodium_crypto_box_seal($signed, $encryptKey),
      ]);

      return 200 == $response->getStatusCode();
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['signingKey'] = [
      '#type' => 'key_select',
      '#title' => $this->t('Signing keys'),
      '#required' => TRUE,
      '#key_filters' => ['type' => 'server_beacon_signkey'],
      '#default_value' => $this->configuration['signingKey'],
      '#description' => $this->t('<p>A server beacon signing keypair of "@key_type". Public key is shared with the reporting station, so transmissions can be verified.</p>', [
        '@key_type' => $this->t('Generated signing keypair'),
      ]),
    ];
    $form['stationPublicKey'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Report transmission public key (base64)'),
      '#max_size' => 40,
      '#required' => TRUE,
      '#pattern' => '^[-A-Za-z0-9+/]*={0,3}$',
      '#default_value' => $this->configuration['stationPublicKey'],
      '#description' => $this->t('Public key used to encrypt transmissions as a base64 encoded string. Key should be provided by the report station.'),
    ];

    return $form;
  }

  /**
   * Get the secret key from the signing keypair.
   *
   * The stored key is a generated sodium keypair, which has both the public
   * and the secret key stored together. This method returns the secret key.
   *
   * @return string|null
   *   Get the private secret key used to sign messages.
   */
  protected function getSigningKey(): ?string {
    $keyId = $this->configuration['signingKey'] ?? NULL;
    if ($keyId && ($key = $this->keyRepository->getKey($keyId))) {
      return \sodium_crypto_sign_secretkey($key->getKeyValue());
    }

    return NULL;
  }

  /**
   * Get the public key for the report station. Used to encrypt the report.
   *
   * The report station will have the private key to decrypt with. The beacons
   * should never have access to the private key.
   *
   * @return string|null
   *   The report station public key for encrypting transmissions.
   */
  protected function getPublicEncryptKey(): ?string {
    if (!empty($this->configuration['stationPublicKey'])) {
      return \sodium_base642bin($this->configuration['stationPublicKey'], SODIUM_BASE64_VARIANT_ORIGINAL);
    }

    return NULL;
  }

}
