<?php

namespace Drupal\server_beacon\Plugin\ServerBeacon\Transmitter;

use Drupal\Component\Plugin\Definition\PluginDefinitionInterface;
use Drupal\Component\Serialization\SerializationInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\key\KeyRepositoryInterface;
use Drupal\server_beacon\Attribute\ServerBeaconTransmitter;
use Drupal\server_beacon\Entity\ServerBeaconReportInterface;
use Drupal\server_beacon\Plugin\BeaconTransmitterInterface;
use Drupal\server_beacon\Plugin\ConfigurablePluginBase;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\RequestOptions;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Synchronous encryption transmission using libsodium.
 *
 * First version of the synchronous server beacon transmitter using the
 * libsodium library.
 *
 * Sodium also supports an asynchronous encryption, which can be coupled with
 * HMAC-SHA256 signing for signature verification. Use the "ASalt1" transmitter
 * for to use the asynchronous key and signing transmission.
 */
#[ServerBeaconTransmitter(
  id: 'salty_v1',
  label: new TranslatableMarkup('Secret key encryption (v1.0)'),
  description: new TranslatableMarkup('Uses the same encryption key on the beacon and report station for encrypting and decrypting report transmissions.'),
)]
class Salty1 extends ConfigurablePluginBase implements BeaconTransmitterInterface, ContainerFactoryPluginInterface {

  /**
   * Create a new instance of the Salty v1 transmitter.
   *
   * @param array $configuration
   *   The configuration settings.
   * @param string $plugin_id
   *   The transmitter plugin identifier.
   * @param array|\Drupal\Component\Plugin\Definition\PluginDefinitionInterface $plugin_definition
   *   The plugin definition.
   * @param \Drupal\Component\Serialization\SerializationInterface $serializer
   *   The data serializer for encoding the data to a encryptable string.
   * @param \Drupal\key\KeyRepositoryInterface $keyRepository
   *   The encryption and signing key manager.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    array|PluginDefinitionInterface $plugin_definition,
    protected SerializationInterface $serializer,
    protected KeyRepositoryInterface $keyRepository,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('serialization.json'),
      $container->get('key.repository')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return ['secretKey' => NULL];
  }

  /**
   * {@inheritdoc}
   */
  public function __invoke(ServerBeaconReportInterface $report, ClientInterface $client, array $data): bool {
    $nonce = random_bytes(SODIUM_CRYPTO_BOX_NONCEBYTES);

    if ($encryptKey = $this->getSecretKey()) {
      // Transmit encrypted status report to the service station.
      $response = $client->request('POST', $report->reportTo(), [
        RequestOptions::HEADERS => [
          'Content-Type' => 'application/octet-stream',
          'Nonce' => \sodium_bin2base64($nonce, SODIUM_BASE64_VARIANT_URLSAFE),
        ],
        RequestOptions::ALLOW_REDIRECTS => FALSE,
        RequestOptions::BODY => \sodium_crypto_secretbox(
          $this->serializer->encode($data),
          $nonce,
          $encryptKey
        ),
      ]);

      return 200 == $response->getStatusCode();
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['secretKey'] = [
      '#type' => 'key_select',
      '#title' => $this->t('Secret key'),
      '#required' => TRUE,
      '#key_filters' => ['type' => 'encryption'],
      '#default_value' => $this->configuration['secretKey'],
      '#description' => $this->t('Secret key that is shared with the report station and used to encrypt transmission. Avoid sharing keys for different servers.'),
    ];

    return $form;
  }

  /**
   * Get the byte value of the encryption key.
   *
   * @return string|null
   *   If the encryption key can be fetched, the encryption key as a byte-string
   *   or NULL if the key can't be loaded.
   */
  protected function getSecretKey(): ?string {
    $keyId = $this->configuration['secretKey'] ?? NULL;
    if ($keyId && ($key = $this->keyRepository->getKey($keyId))) {
      return $key->getKeyValue();
    }

    return NULL;
  }

}
