<?php

namespace Drupal\server_beacon;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Config\Entity\ConfigEntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\server_beacon\Entity\ServerBeaconReportInterface;
use GuzzleHttp\ClientInterface;

/**
 * Service class for facilitating the transmission of reports.
 *
 * Generates and transmits server beacon reports to a designated service
 * station.
 */
class Reporter implements ReporterInterface {

  /**
   * The server beacon report entity storage handler.
   *
   * @var \Drupal\Core\Config\Entity\ConfigEntityStorageInterface
   */
  protected ConfigEntityStorageInterface $reportStorage;

  /**
   * Creates a new instance of the Reporter service class.
   *
   * @param \GuzzleHttp\ClientInterface $client
   *   The HTTP client instance for transmitting data over HTTPS.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The persistent lock service to ensure only one instance runs on
   *   multi-node applications.
   * @param \Drupal\Component\Uuid\UuidInterface $uuidGenerator
   *   The UUID generator service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The system time service for getting request time.
   */
  public function __construct(
    protected ClientInterface $client,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected LockBackendInterface $lock,
    protected UuidInterface $uuidGenerator,
    protected TimeInterface $time,
  ) {
    /** @var \Drupal\Core\Config\Entity\ConfigEntityStorageInterface $storage */
    $storage = $entityTypeManager->getStorage('server_beacon_report');
    $this->reportStorage = $storage;
  }

  /**
   * {@inheritdoc}
   */
  public function generateReport(string|ServerBeaconReportInterface $report, string $environment = ''): array {
    $report = $this->resolveReport($report);

    // Ensure that this data is included as part of the encrypted payload or
    // message digest signature to make sure it hasn't been tampered with.
    $data = [];
    foreach ($report->getBeacons() as $beacon) {
      $data[$beacon->getPluginId()] = $beacon->fetchStatus();
    }

    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function transmitReport(string|ServerBeaconReportInterface $report, string $environment = ''): bool {
    $report = $this->resolveReport($report);

    if ($transmitter = $report->getTransmitter()) {
      $lockId = 'server_beacon:' . $report->id();

      try {
        if ($this->lock->acquire($lockId, 300) && ($data = $this->generateReport($report, $environment))) {
          return $transmitter($report, $this->client, [
            'name' => $report->id(),
            'environment' => $environment,
            'uuid' => $this->uuidGenerator->generate(),
            'created' => $this->time->getRequestTime(),
            'data' => $data,
          ]);
        }
      }
      finally {
        // Ensure the lock is released when the report transmission is complete.
        $this->lock->release($lockId);
      }
    }

    return FALSE;
  }

  /**
   * Ensure that the server beacon report variable is a load report object.
   *
   * @param string|\Drupal\server_beacon\Entity\ServerBeaconReportInterface $report
   *   Either the ID of the report to load, or the server report object.
   *
   * @return \Drupal\server_beacon\Entity\ServerBeaconReportInterface
   *   The loaded report config entity.
   */
  protected function resolveReport(string|ServerBeaconReportInterface $report): ServerBeaconReportInterface {
    if ($report instanceof ServerBeaconReportInterface) {
      return $report;
    }
    /** @var \Drupal\server_beacon\Entity\ServerBeaconReportInterface $object */
    $object = $this->reportStorage->load($report);
    if ($object) {
      return $object;
    }

    throw new \InvalidArgumentException(sprintf('Unable to find report for: %s', $report));
  }

}
